# -*- coding: utf-8 -*-

"""Data Transformation Graph."""

__author__ = "Nathaniel Starkman"
__credits__ = ["Astropy"]


__all__ = [
    "TransformGraph",
]


##############################################################################
# IMPORTS

# BUILT-IN
import heapq
import typing as T
from collections import defaultdict

# PROJECT-SPECIFIC
from .transformations import CompositeTransform, _default_xfm_set
from utilipy.utils import functools, inspect

##############################################################################
# PARAMETERS


##############################################################################
# CODE
##############################################################################


class TransformGraph:
    """Graph representing the paths between data types.

    Notes
    -----
    Note that the _graph key-value order is "totype": "fromtype".
    This is the opposite type as Astropy's since this TransformGraph
    needs to support:

    - catch-all conversions, such as converting from any sequence ``data``
      to type ``list`` by applying ``list(data)``
    - sub-type conversions, where the conversion a -> b (with types A, B)
      works on any subtype of A (like A1(A)).


    .. todo::

        - catch-all conversions to a specific type
        - multiple input option conversions to a specific type
        - multiple output option conversions by choosing one with shortest path
        - pre-register basic conversions like None->None, list->tuple

    """

    def __init__(self, seed_basic: bool = True):
        """Data Transformation Graph.

        Parameters
        ----------
        seed_basic : bool
            whether to start with a basic set of transformations

            .. todo::

                Generate documentation / colored graph like Astropy's

        """
        # graph, in reverse order
        self._graph = _default_xfm_set if seed_basic else defaultdict(dict)
        self.invalidate_cache()  # generates cache entries

    # /def

    @property
    def _cached_names(self):
        if self._cached_names_dct is None:
            self._cached_names_dct = dct = {}
            for c in self.type_set:
                nm = getattr(c, "name", None)
                if nm is not None:
                    if not isinstance(nm, list):
                        nm = [nm]
                    for name in nm:
                        dct[name] = c

        return self._cached_names_dct

    # /def

    @property
    def type_set(self):
        """A `set` of all data types present in this `TransformGraph`."""
        if self._cached_type_set is None:
            self._cached_type_set = set()
            for a in self._graph:
                self._cached_type_set.add(a)
                for b in self._graph[a]:
                    self._cached_type_set.add(b)

        return self._cached_type_set.copy()

    # /def

    def invalidate_cache(self):
        """Clears all caching attributes.

        Invalidates the cache that stores optimizations for traversing the
        transform graph.  This is called automatically when transforms
        are added or removed, but will need to be called manually if
        weights on transforms are modified inplace.

        """
        self._cached_names_dct: T.Optional[dict] = None
        self._cached_type_set: T.Optional[set] = None
        # self._cached_frame_attributes = None
        # self._cached_component_names = None
        self._shortestpaths = {}
        self._composite_cache = {}

    # /def

    def add_transform(self, fromtype, totype, transform):
        """Add a new data transformation to the graph.

        .. todo::

           - support an "Any" option in fromtype
           - support adding a tuple of types as the "fromtype"
           - support subtypes in "fromtype"

        Parameters
        ----------
        fromtype : class
            The data class to start from.
        totype : class
            The data class to transform into.
        transform : DataTransform or similar callable
            The transformation object. Typically a `DataTransform` object,
            although it may be some other callable that is called with the same
            signature.

        Raises
        ------
        TypeError
            If ``fromtype`` or ``totype`` are not classes or ``transform`` is
            not callable.

        """
        if not inspect.isclass(fromtype) and fromtype is not None:
            raise TypeError("fromtype must be a class")
        if not inspect.isclass(totype) and totype is not None:
            raise TypeError("totype must be a class")
        if not callable(transform):
            raise TypeError("transform must be callable")

        type_set = self.type_set.copy()
        type_set.add(fromtype)
        type_set.add(totype)

        self._graph[totype][fromtype] = transform
        self.invalidate_cache()

    # /def

    def remove_transform(self, fromtype, totype, transform):
        """Removes a data transform from the graph.

        .. todo::

            - support removing catch-all transformations

        Parameters
        ----------
        fromtype : class or `None`
            The coordinate frame *class* to start from. If `None`,
            ``transform`` will be searched for and removed (``totype`` must
            also be `None`).
        totype : class or `None`
            The coordinate frame *class* to transform into. If `None`,
            ``transform`` will be searched for and removed (``fromtype`` must
            also be `None`).
        transform : callable or `None`
            The transformation object to be removed or `None`.  If `None`
            and ``totype`` and ``fromtype`` are supplied, there will be no
            check to ensure the correct object is removed.

        """
        if fromtype is None or totype is None:
            if not (totype is None and fromtype is None):
                raise ValueError(
                    "fromtype and totype must both be None if either are"
                )
            if transform is None:
                raise ValueError("cannot give all Nones to remove_transform")

            # search for the requested transform by brute force and remove it
            for a in self._graph:
                agraph = self._graph[a]
                for b in agraph:
                    if agraph[b] is transform:
                        del agraph[b]
                        fromtype = a
                        break

                # If transform was found, need to break out of outer loop
                if fromtype:
                    break
            else:
                raise ValueError(
                    "Could not find transform {} in the "
                    "graph".format(transform)
                )

        else:
            if transform is None:
                self._graph[totype].pop(fromtype, None)
            else:
                curr = self._graph[totype].get(fromtype, None)
                if curr is transform:
                    self._graph[totype].pop(fromtype)
                else:
                    raise ValueError(
                        "Current transform from {} to {} is not "
                        "{}".format(fromtype, totype, transform)
                    )

        # Remove the subgraph if it is now empty
        if self._graph[totype] == {}:
            self._graph.pop(totype)

        self.invalidate_cache()

    # /def

    def _construct_path(self, fromtype, totype):
        """Construct path using Dijkstra's algorithm.

        Parameters
        ----------
        fromtype
        totype

        Returns
        -------
        path : list
        priority : int

        """
        inf = float("inf")

        nodes = []
        # first make the list of nodes
        for a in self._graph:
            if a not in nodes:
                nodes.append(a)
            for b in self._graph[a]:
                if b not in nodes:
                    nodes.append(b)

        if fromtype not in nodes or totype not in nodes:
            # fromtype or totype are isolated or not registered, so there's
            # certainly no way to get from one to the other
            return None, inf

        edgeweights = {}
        # construct another graph that is a dict of dicts of priorities
        # (used as edge weights in Dijkstra's algorithm)
        for a in self._graph:
            edgeweights[a] = aew = {}
            agraph = self._graph[a]
            for b in agraph:
                aew[b] = float(
                    agraph[b].priority if hasattr(agraph[b], "priority") else 1
                )

        # entries in q are [distance, count, nodeobj, pathlist]
        # count is needed because in py 3.x, tie-breaking fails on the nodes.
        # this way, insertion order is preserved if the weights are the same
        # q = [[inf,i,n,[]] for i, n in enumerate(nodes) if n is not fromtype]
        # q.insert(0, [0, -1, fromtype, []])
        q = [[inf, i, n, []] for i, n in enumerate(nodes) if n is not totype]
        q.insert(0, [0, -1, totype, []])

        # this dict stores the distance to node from ``fromtype`` and the path
        result = {}

        # definitely starts as a valid heap because of the insert line;
        # from the node to itself is always the shortest distance
        while len(q) > 0:
            d, orderi, n, path = heapq.heappop(q)

            if d == inf:
                # everything left is unreachable from fromtype,
                # just copy them to the results and jump out of the loop
                result[n] = (None, d)
                for d, orderi, n, path in q:
                    result[n] = (None, d)
                break
            else:
                result[n] = (path, d)
                path.append(n)
                if n not in edgeweights:
                    # a system that can be transformed to, but not from.
                    continue
                for n2 in edgeweights[n]:
                    if n2 not in result:  # already visited
                        # find where n2 is in the heap
                        for i in range(len(q)):
                            if q[i][2] == n2:
                                break
                        else:
                            raise ValueError(
                                "n2 not in heap - this should be impossible!"
                            )

                        newd = d + edgeweights[n][n2]
                        if newd < q[i][0]:
                            q[i][0] = newd
                            q[i][3] = list(path)
                            heapq.heapify(q)

        # cache for later use
        # self._shortestpaths[totype] = result # FIXME
        # return result[fromtype]
        # FIXME
        path, d = result[fromtype]
        return path[::-1], d

    # /def

    def find_shortest_path(self, fromtype, totype):
        """Compute shortest path along graph from one system to another.

        Parameters
        ----------
        fromtype : class
            The coordinate frame class to start from.
        totype : class
            The coordinate frame class to transform into.

        Returns
        -------
        path : list of classes or `None`
            The path from ``fromtype`` to ``totype`` as an in-order sequence
            of classes.  This list includes *both* ``fromtype`` and
            ``totype``. Is `None` if there is no possible path.
        distance : number
            The total distance/priority from ``fromtype`` to ``totype``.  If
            priorities are not set this is the number of transforms
            needed. Is ``inf`` if there is no possible path.

        """
        # ----------------------------------
        # special-case the 0 or 1-path

        if totype is fromtype:
            if fromtype not in self._graph[totype]:
                # Means there's no transform necessary to go from it to itself.
                return [totype], 0

        if fromtype in self._graph[totype]:
            # this will also catch the case where totype is fromtype, but has
            # a defined transform.
            t = self._graph[totype][fromtype]
            return (
                [fromtype, totype],
                float(t.priority if hasattr(t, "priority") else 1),
            )

        # ----------------------------------
        # otherwise, need to construct the path:

        inf = float("inf")

        # TODO verify this works for catch-alls
        if totype in self._shortestpaths:
            # already have a cached result
            fpaths = self._shortestpaths[totype]
            if fromtype in fpaths:
                return fpaths[fromtype]
            else:
                path, priority = None, inf

        path, priority = self._construct_path(fromtype, totype)

        return path, priority

    # /def

    def get_transform(self, fromtype, totype):
        """Generate `CompositeTransform` for a datatype transformation.

        Parameters
        ----------
        fromtype : class
            The coordinate frame class to start from.
        totype : class
            The coordinate frame class to transform into.

        Returns
        -------
        trans : `CompositeTransform` or `None`
            If there is a path from ``fromtype`` to ``totype``, this is a
            transform object for that path. If no path could be found, this is
            `None`.

        Notes
        -----
        This function always returns a `CompositeTransform`, because
        `CompositeTransform` is slightly more adaptable in the way it can be
        called than other transform classes. Specifically, it takes care of
        intermediate steps of transformations in a way that is consistent with
        1-hop transformations.

        """
        if not inspect.isclass(fromtype) and fromtype is not None:
            raise TypeError("fromtype is not a class")
        if not inspect.isclass(totype) and totype is not None:
            raise TypeError("totype is not a class")

        path, distance = self.find_shortest_path(fromtype, totype)

        if path is None:
            return None

        transforms = []
        currtype = path[0]
        for p in path[1:]:  # first element is fromtype so we skip it
            transforms.append(self._graph[p][currtype])
            currtype = p

        fttuple = (fromtype, totype)
        if fttuple not in self._composite_cache:
            comptrans = CompositeTransform(
                transforms, fromtype, totype, register_graph=False
            )
            self._composite_cache[fttuple] = comptrans

        return self._composite_cache[fttuple]

    # /def

    def lookup_name(self, name: str):
        """Tries to locate the class with the provided alias.

        Parameters
        ----------
        name : str
            The alias to look up.

        Returns
        -------
        datacls
            The data class corresponding to the ``name`` or `None` if
            no such class exists.

        """
        return self._cached_names.get(name, None)

    # /def

    def get_names(self):
        """Returns all available transform names.

        Returns
        -------
        nms : list
            The aliases for coordinate systems.
            They will all be valid arguments to `lookup_name`.

        """
        return list(self._cached_names.keys())

    # /def

    def register(
        self, transcls, fromtype, totype, priority: int = 1, **kwargs
    ):
        """A function decorator for defining a transformation.

        .. note::
            If decorating a static method of a class, ``@staticmethod``
            should be  added *above* this decorator.

        Parameters
        ----------
        transcls : class
            The class of the transformation object to create.
        fromtype : class
            The data class to start from.
        totype : class
            The data class to transform into.
        priority : number
            The priority if this transform when finding the shortest
            coordinate transform path - large numbers are lower priorities.

        Additional keyword arguments are passed into the ``transcls``
        constructor.

        Returns
        -------
        deco : function
            A function that can be called on another function as a decorator
            (see example).

        Notes
        -----
        This decorator assumes the first argument of the ``transcls``
        initializer accepts a callable, and that the second and third
        are ``fromtype`` and ``totype``. If this is not true, you should just
        initialize the class manually and use `add_transform` instead of
        using this decorator.

        Examples
        --------
        ::

            graph = TransformGraph()

            @graph.transform(DataTransform, list, tuple)
            def list_to_tuple(data):
                return tuple(data)

        """
        # create decorator
        def register_xfm_decorator(func: T.Callable):
            # this doesn't do anything directly with the transform because
            # ``register_graph=self`` stores it in the transform graph
            # automatically
            transcls(
                func,
                fromtype,
                totype,
                priority=priority,
                register_graph=self,
                **kwargs,
            )
            return func

        # /def

        return register_xfm_decorator

    # /def

    transform = register  # for similarity to Astropy's
    # /def

    def function_decorator(
        self,
        function: T.Optional[T.Callable] = None,
        *,
        _doc_style: str = "numpy",
        _doc_fmt: T.Dict[str, T.Any] = {},
        **arguments,
    ):
        """Apply data transformations to function arguments.

        Parameters
        ----------
        function : T.Callable or None, optional
            the function to be decoratored
            if None, then returns decorator to apply.

        **arguments: dict
            argument information, where keyword is the argument parameter
            name in `function`. The values are either the desired output type
            or a 3-element *tuple* in the following order
            (outtype, (args), dict(kwargs)). The args and kwargs are passed
            into the transformation.

        Returns
        -------
        wrapper : T.Callable
            wrapper for `function` that manage input catalog tables.
            includes the original function in a method `.__wrapped__`

        Other Parameters
        ----------------
        _doc_style: str or formatter, optional
            `function` docstring style. Parameter to `wraps`.
        _doc_fmt: dict, optional
            `function` docstring format arguments. Parameter to `wraps`.

        Notes
        -----
        .. todo::

            - scrape output type from function argument annotation
            - support a multiple possible output types (Union[etc]), choosing
              the one with the shortest path

        """
        if function is None:  # allowing for optional arguments
            return functools.partial(
                self.function_decorator,
                _doc_style=_doc_style,
                _doc_fmt=_doc_fmt,
                **arguments,
            )

        sig = inspect.fuller_signature(function)
        _doc_fmt.update({"argkeys": ", ".join(arguments.keys())})

        @functools.wraps(function, _doc_style=_doc_style, _doc_fmt=_doc_fmt)
        def wrapper(*args, _skip_decorator=False, **kwargs):
            """Wrapper docstring.

            Other Parameters
            ----------------
            _skip_decorator : bool, optional
                Whether to skip the decorator.
                default {_skip_decorator}

            Notes
            -----
            This function is wrapped with a data `~TransformGraph` decorator.
            See `~TransformGraph.function_decorator` for details.
            The transformation arguments are also attached to this function
            as the attribute ``._transforms``.
            The affected arguments are: {argkeys}

            """
            if _skip_decorator:  # whether to skip decorator or keep going
                return function(*args, **kwargs)
            # else:

            ba = sig.bind_partial_with_defaults(*args, **kwargs)

            for name, outtype in arguments.items():

                data = ba.arguments[name]  # get the data to be transformed

                # The values are either the desired output type
                # or a 3-element *tuple* in the following order
                # (outtype, (args), dict(kwargs)). The args and kwargs
                # are passed into the transformation.
                t_args, t_kw = (), {}  # assume no input (kw)args
                if isinstance(outtype, tuple):  # output type or info tuple
                    if len(outtype) == 3:  # it's an info tuple
                        outtype, t_args, t_kw = outtype

                # get transformation
                fromtype = type(data) if data is not None else None
                t = self.get_transform(fromtype, outtype)
                # print(t, type(data), outtype)
                # apply transformation
                ba.arguments[name] = t(data, *t_args, **t_kw)

            # /def

            return_ = function(*ba.args, **ba.kwargs)

            return return_

        # /def

        wrapper._transforms = arguments

        return wrapper

        # /def

    decorate = function_decorator
    # /def

    # def copy(self):
    #     """Deep-copy self"""
    #     return copy.deepcopy(self)

    # # /def


# /class


##############################################################################
# END



# -*- coding: utf-8 -*-

# ----------------------------------------------------------------------------
#
# TITLE   : extended base imports
# PROJECT : utilipy
#
# ----------------------------------------------------------------------------


r"""Extending the standard import file.

Routine Listings
----------------
Numpy: imports

    - linalg.norm

Scipy: imports

    - stats.binned_statistic->binned_stats

References
----------
SciPy references are [#]_ and [#]_.
NumPy references are [#]_ and [#]_.
IPython reference is [#]_.
Matplotlib reference is [#]_.

.. [#] Travis E. Oliphant. Python for Scientific Computing, Computing in
    Science and Engineering, 9, 10-20 (2007), DOI:10.1109/MCSE.2007.58
    http://scitation.aip.org/content/aip/journal/cise/9/3/10.1109/MCSE.2007.58

.. [#] K. Jarrod Millman and Michael Aivazis. Python for Scientists and
    Engineers, Computing in Science and Engineering, 13, 9-12 (2011),
    DOI:10.1109/MCSE.2011.36
    http://scitation.aip.org/content/aip/journal/cise/13/2/10.1109/MCSE.2011.36

.. [#] Travis E, Oliphant. A guide to NumPy, USA: Trelgol Publishing, (2006).

.. [#] Stéfan van der Walt, S. Chris Colbert and Gaël Varoquaux.
    The NumPy Array: A Structure for Efficient Numerical Computation,
    Computing in Science and Engineering, 13, 22-30 (2011),
    DOI:10.1109/MCSE.2011.37
    http://scitation.aip.org/content/aip/journal/cise/13/2/10.1109/MCSE.2011.37

.. [#] Fernando Pérez, Brian E. Granger, IPython: A System for Interactive
    Scientific Computing, Computing in Science and Engineering, vol. 9,
    no. 3, pp. 21-29, May/June 2007, doi:10.1109/MCSE.2007.53.
    URL: https://ipython.org

.. [#] John D. Hunter. Matplotlib: A 2D Graphics Environment, Computing in
    Science and Engineering, 9, 90-95 (2007), DOI:10.1109/MCSE.2007.55
    http://scitation.aip.org/content/aip/journal/cise/9/3/10.1109/MCSE.2007.55


"""

__author__ = "Nathaniel Starkman"


__all__ = [
    # functions
    "extended_imports_help",
    # imports
    "norm",
    "binned_stats",
]


##############################################################################
# HELPER FUNCTIONS

# THIRD PARTY
from numpy.linalg import norm
from scipy.stats import binned_statistic as binned_stats

# PROJECT-SPECIFIC
from utilipy.imports import conf
from utilipy.utils import make_help_function

##############################################################################
# IMPORTS


##############################################################################
# Printing Information

extended_imports_help = make_help_function(
    "extend", __doc__, look_for="Routine Listings"
)


if conf.verbose_imports:
    extended_imports_help()


##############################################################################
# END



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# ----------------------------------------------------------------------------
#
# TITLE   : test_select
#
# ----------------------------------------------------------------------------


"""Test functions for data_utils/select."""

__author__ = "Nathaniel Starkman"


##############################################################################
# IMPORTS

# THIRD PARTY
import numpy as np
import pytest

# PROJECT-SPECIFIC
from utilipy.data_utils.select import _inRange, ioRange, outRange

from utilipy.data_utils.select import inRange  # ellipse,; circle,


##############################################################################
# PARAMETERS

x = np.arange(2)  # [0, 1]
y = np.arange(2) + 10  # [10, 11]

z = np.c_[x, y].T  # [[0,  1], [10, 11]]
zT = z.T  # [[0,  10], [1, 11]]


##############################################################################
# _inRange


def test__inRange_1D_default():
    """Test _inRange."""
    # standard  (lbi=True, ubi=False)
    assert all(_inRange(x, rng=[0, 1]) == np.array([True, False]))


# /def


@pytest.mark.parametrize(
    "flags, expected",
    [
        ((False, False), np.array([False, False])),
        ((False, True), np.array([False, True])),
        ((True, False), np.array([True, False])),
        ((True, True), np.array([True, True])),
    ],
    ids=["non-inclusive", "ubi", "lbi", "inclusive"],
)
def test__inRange_1D_options(flags, expected):
    """Test _inRange, 1D, flag options."""
    assert all(_inRange(x, rng=[0, 1], lbi=flags[0], ubi=flags[1]) == expected)


# /def


def test__inRange():
    """Test _inRange."""

    # ---------------------------------
    # N-Dimensional

    rng = [[0, 1], [10, 11]]

    # standard  (lbi=True, ubi=False)
    assert (_inRange(z, rng) == np.array([[True, False], [True, False]])).all()

    # going through options
    assert (
        _inRange(z, rng, lbi=False, ubi=False)
        == np.array([[False, False], [False, False]])
    ).all()
    assert (
        _inRange(z, rng, lbi=False, ubi=True)
        == np.array([[False, True], [False, True]])
    ).all()
    assert (
        _inRange(z, rng, lbi=True, ubi=False)
        == np.array([[True, False], [True, False]])
    ).all()
    assert (
        _inRange(z, rng, lbi=True, ubi=True)
        == np.array([[True, True], [True, True]])
    ).all()

    return


# /def


##############################################################################
# inRange


def test_inRange_single_1D_argument():
    """Test inRange for a single, 1D argument."""
    # set range
    rng = [0, 1]

    # standard  (lbi=True, ubi=False)
    assert all(inRange(x, rng=rng) == np.array([True, False]))

    # going through options
    assert all(
        inRange(x, rng=rng, lbi=False, ubi=False) == np.array([False, False])
    )
    assert all(
        inRange(x, rng=rng, lbi=False, ubi=True) == np.array([False, True])
    )
    assert all(
        inRange(x, rng=rng, lbi=True, ubi=False) == np.array([True, False])
    )
    assert all(
        inRange(x, rng=rng, lbi=True, ubi=True) == np.array([True, True])
    )

    return


# /def


def test_inRange_many_1D_argmuents():
    """Test inRange for many 1D arguments."""
    # ---------------------------------
    # range agreement
    # set range
    rng = [[0, 1], [10, 11]]

    # standard  (lbi=True, ubi=False)
    assert (inRange(x, y, rng=rng) == np.array([True, False])).all()

    # going through options
    assert (
        inRange(x, y, rng=rng, lbi=False, ubi=False)
        == np.array([False, False])
    ).all()
    assert (
        inRange(x, y, rng=rng, lbi=False, ubi=True) == np.array([False, True])
    ).all()
    assert (
        inRange(x, y, rng=rng, lbi=True, ubi=False) == np.array([True, False])
    ).all()
    assert (
        inRange(x, y, rng=rng, lbi=True, ubi=True) == np.array([True, True])
    ).all()

    # ---------------------------------
    # different ranges, the second never matters

    rng = [[0, 1], [9, 12]]

    # standard  (lbi=True, ubi=False)
    assert (inRange(x, y, rng=rng) == np.array([True, False])).all()

    # going through options
    assert (
        inRange(x, y, rng=rng, lbi=False, ubi=False)
        == np.array([False, False])
    ).all()
    assert (
        inRange(x, y, rng=rng, lbi=False, ubi=True) == np.array([False, True])
    ).all()
    assert (
        inRange(x, y, rng=rng, lbi=True, ubi=False) == np.array([True, False])
    ).all()
    assert (
        inRange(x, y, rng=rng, lbi=True, ubi=True) == np.array([True, True])
    ).all()

    return


# /def


def test_inRange_single_ND_argument():
    """Test inRange for a single N-D argument."""
    # ---------------------------------
    # range agreement

    rng = [[0, 1], [10, 11]]

    # standard  (lbi=True, ubi=False)
    assert (
        inRange(z, rng=rng) == np.array([[True, False], [True, False]])
    ).all()

    # going through options
    assert (
        inRange(z, rng=rng, lbi=False, ubi=False)
        == np.array([[False, False], [False, False]])
    ).all()
    assert (
        inRange(z, rng=rng, lbi=False, ubi=True)
        == np.array([[False, True], [False, True]])
    ).all()
    assert (
        inRange(z, rng=rng, lbi=True, ubi=False)
        == np.array([[True, False], [True, False]])
    ).all()
    assert (
        inRange(z, rng=rng, lbi=True, ubi=True)
        == np.array([[True, True], [True, True]])
    ).all()

    # ---------------------------------
    # different ranges

    rng = [[0, 1], [9, 12]]

    # standard  (lbi=True, ubi=False)
    assert (
        inRange(z, rng=rng) == np.array([[True, False], [True, True]])
    ).all()

    # going through options
    assert (
        inRange(z, rng=rng, lbi=False, ubi=False)
        == np.array([[False, False], [True, True]])
    ).all()
    assert (
        inRange(z, rng=rng, lbi=False, ubi=True)
        == np.array([[False, True], [True, True]])
    ).all()
    assert (
        inRange(z, rng=rng, lbi=True, ubi=False)
        == np.array([[True, False], [True, True]])
    ).all()
    assert (
        inRange(z, rng=rng, lbi=True, ubi=True)
        == np.array([[True, True], [True, True]])
    ).all()

    return


# /def


def test_inRange_many_ND_arguments():
    """Test inRange for many N-D arguments."""
    # ---------------------------------
    # range agreement

    rng = ([[0, 1], [10, 11]], [[0, 10], [1, 11]])  # for z  # for zT

    # standard  (lbi=True, ubi=False)
    assert (
        inRange(z, zT, rng=rng) == np.array([[True, False], [True, False]])
    ).all()

    # going through options
    assert (
        inRange(z, zT, rng=rng, lbi=False, ubi=False)
        == np.array([[False, False], [False, False]])
    ).all()
    assert (
        inRange(z, zT, rng=rng, lbi=False, ubi=True)
        == np.array([[False, True], [False, True]])
    ).all()
    assert (
        inRange(z, zT, rng=rng, lbi=True, ubi=False)
        == np.array([[True, False], [True, False]])
    ).all()
    assert (
        inRange(z, zT, rng=rng, lbi=True, ubi=True)
        == np.array([[True, True], [True, True]])
    ).all()

    # ---------------------------------
    # different ranges, z rng does not matter

    rng = ([[-1, 2], [9, 12]], [[0, 10], [1, 11]])  # for z  # for zT

    # standard  (lbi=True, ubi=False)
    assert (
        inRange(z, zT, rng=rng) == np.array([[True, False], [True, False]])
    ).all()

    # going through options
    assert (
        inRange(z, zT, rng=rng, lbi=False, ubi=False)
        == np.array([[False, False], [False, False]])
    ).all()
    assert (
        inRange(z, zT, rng=rng, lbi=False, ubi=True)
        == np.array([[False, True], [False, True]])
    ).all()
    assert (
        inRange(z, zT, rng=rng, lbi=True, ubi=False)
        == np.array([[True, False], [True, False]])
    ).all()
    assert (
        inRange(z, zT, rng=rng, lbi=True, ubi=True)
        == np.array([[True, True], [True, True]])
    ).all()

    # ---------------------------------
    # different ranges, zT rng does not matter

    rng = ([[0, 1], [10, 11]], [[-1, 11], [0, 12]])  # for z  # for zT

    # standard  (lbi=True, ubi=False)
    assert (
        inRange(z, zT, rng=rng) == np.array([[True, False], [True, False]])
    ).all()

    # going through options
    assert (
        inRange(z, zT, rng=rng, lbi=False, ubi=False)
        == np.array([[False, False], [False, False]])
    ).all()
    assert (
        inRange(z, zT, rng=rng, lbi=False, ubi=True)
        == np.array([[False, True], [False, True]])
    ).all()
    assert (
        inRange(z, zT, rng=rng, lbi=True, ubi=False)
        == np.array([[True, False], [True, False]])
    ).all()
    assert (
        inRange(z, zT, rng=rng, lbi=True, ubi=True)
        == np.array([[True, True], [True, True]])
    ).all()

    return


# /def


def test_inRange_mixing_1D_and_ND_arguments():
    """Test inRange for a mix of 1D and N-D arguments.

    only mixed by breaking the ND arguments into a series of 1D arrays

    """
    # ---------------------------------
    # range agreement

    rng = (
        [0, 1],  # for x
        *[[0, 1], [10, 11]],  # for z
        [10, 11],  # for y
        *[[0, 10], [1, 11]],  # for zT
    )

    # standard  (lbi=True, ubi=False)
    assert (
        inRange(x, *z, y, *zT, rng=rng)
        == np.array([[True, False], [True, False]])
    ).all()

    # going through options
    assert (
        inRange(x, *z, y, *zT, rng=rng, lbi=False, ubi=False)
        == np.array([[False, False], [False, False]])
    ).all()
    assert (
        inRange(x, *z, y, *zT, rng=rng, lbi=False, ubi=True)
        == np.array([[False, True], [False, True]])
    ).all()
    assert (
        inRange(x, *z, y, *zT, rng=rng, lbi=True, ubi=False)
        == np.array([[True, False], [True, False]])
    ).all()
    assert (
        inRange(x, *z, y, *zT, rng=rng, lbi=True, ubi=True)
        == np.array([[True, True], [True, True]])
    ).all()

    return


# /def


##############################################################################
# outRange


def test_outRange_single_1D_argument():
    """Test outRange for a single, 1D argument."""
    # set range
    rng = [0, 1]

    # standard  (lbi=True, ubi=False)
    assert all(outRange(x, rng=rng) == ~np.array([True, False]))

    # going through options
    assert all(
        outRange(x, rng=rng, lbi=False, ubi=False) == ~np.array([False, False])
    )
    assert all(
        outRange(x, rng=rng, lbi=False, ubi=True) == ~np.array([False, True])
    )
    assert all(
        outRange(x, rng=rng, lbi=True, ubi=False) == ~np.array([True, False])
    )
    assert all(
        outRange(x, rng=rng, lbi=True, ubi=True) == ~np.array([True, True])
    )

    return


# /def


def test_outRange_many_1D_argmuents():
    """Test outRange for many 1D arguments."""
    # ---------------------------------
    # range agreement
    # set range
    rng = [[0, 1], [10, 11]]

    # standard  (lbi=True, ubi=False)
    assert (outRange(x, y, rng=rng) == ~np.array([True, False])).all()

    # going through options
    assert (
        outRange(x, y, rng=rng, lbi=False, ubi=False)
        == ~np.array([False, False])
    ).all()
    assert (
        outRange(x, y, rng=rng, lbi=False, ubi=True)
        == ~np.array([False, True])
    ).all()
    assert (
        outRange(x, y, rng=rng, lbi=True, ubi=False)
        == ~np.array([True, False])
    ).all()
    assert (
        outRange(x, y, rng=rng, lbi=True, ubi=True) == ~np.array([True, True])
    ).all()

    # ---------------------------------
    # different ranges, the second never matters

    rng = [[0, 1], [9, 12]]

    # standard  (lbi=True, ubi=False)
    assert (outRange(x, y, rng=rng) == ~np.array([True, False])).all()

    # going through options
    assert (
        outRange(x, y, rng=rng, lbi=False, ubi=False)
        == ~np.array([False, False])
    ).all()
    assert (
        outRange(x, y, rng=rng, lbi=False, ubi=True)
        == ~np.array([False, True])
    ).all()
    assert (
        outRange(x, y, rng=rng, lbi=True, ubi=False)
        == ~np.array([True, False])
    ).all()
    assert (
        outRange(x, y, rng=rng, lbi=True, ubi=True) == ~np.array([True, True])
    ).all()

    return


# /def


def test_outRange_single_ND_argument():
    """Test outRange for a single N-D argument."""
    # ---------------------------------
    # range agreement

    rng = [[0, 1], [10, 11]]

    # standard  (lbi=True, ubi=False)
    assert (
        outRange(z, rng=rng) == ~np.array([[True, False], [True, False]])
    ).all()

    # going through options
    assert (
        outRange(z, rng=rng, lbi=False, ubi=False)
        == ~np.array([[False, False], [False, False]])
    ).all()
    assert (
        outRange(z, rng=rng, lbi=False, ubi=True)
        == ~np.array([[False, True], [False, True]])
    ).all()
    assert (
        outRange(z, rng=rng, lbi=True, ubi=False)
        == ~np.array([[True, False], [True, False]])
    ).all()
    assert (
        outRange(z, rng=rng, lbi=True, ubi=True)
        == ~np.array([[True, True], [True, True]])
    ).all()

    # ---------------------------------
    # different ranges

    rng = [[0, 1], [9, 12]]

    # standard  (lbi=True, ubi=False)
    assert (
        outRange(z, rng=rng) == ~np.array([[True, False], [True, True]])
    ).all()

    # going through options
    assert (
        outRange(z, rng=rng, lbi=False, ubi=False)
        == ~np.array([[False, False], [True, True]])
    ).all()
    assert (
        outRange(z, rng=rng, lbi=False, ubi=True)
        == ~np.array([[False, True], [True, True]])
    ).all()
    assert (
        outRange(z, rng=rng, lbi=True, ubi=False)
        == ~np.array([[True, False], [True, True]])
    ).all()
    assert (
        outRange(z, rng=rng, lbi=True, ubi=True)
        == ~np.array([[True, True], [True, True]])
    ).all()

    return


# /def


def test_outRange_many_ND_arguments():
    """Test outRange for many N-D arguments."""
    # ---------------------------------
    # range agreement

    rng = ([[0, 1], [10, 11]], [[0, 10], [1, 11]])  # for z  # for zT

    # standard  (lbi=True, ubi=False)
    assert (
        outRange(z, zT, rng=rng) == ~np.array([[True, False], [True, False]])
    ).all()

    # going through options
    assert (
        outRange(z, zT, rng=rng, lbi=False, ubi=False)
        == ~np.array([[False, False], [False, False]])
    ).all()
    assert (
        outRange(z, zT, rng=rng, lbi=False, ubi=True)
        == ~np.array([[False, True], [False, True]])
    ).all()
    assert (
        outRange(z, zT, rng=rng, lbi=True, ubi=False)
        == ~np.array([[True, False], [True, False]])
    ).all()
    assert (
        outRange(z, zT, rng=rng, lbi=True, ubi=True)
        == ~np.array([[True, True], [True, True]])
    ).all()

    # ---------------------------------
    # different ranges, z rng does not matter

    rng = ([[-1, 2], [9, 12]], [[0, 10], [1, 11]])  # for z  # for zT

    # standard  (lbi=True, ubi=False)
    assert (
        outRange(z, zT, rng=rng) == ~np.array([[True, False], [True, False]])
    ).all()

    # going through options
    assert (
        outRange(z, zT, rng=rng, lbi=False, ubi=False)
        == ~np.array([[False, False], [False, False]])
    ).all()
    assert (
        outRange(z, zT, rng=rng, lbi=False, ubi=True)
        == ~np.array([[False, True], [False, True]])
    ).all()
    assert (
        outRange(z, zT, rng=rng, lbi=True, ubi=False)
        == ~np.array([[True, False], [True, False]])
    ).all()
    assert (
        outRange(z, zT, rng=rng, lbi=True, ubi=True)
        == ~np.array([[True, True], [True, True]])
    ).all()

    # ---------------------------------
    # different ranges, zT rng does not matter

    rng = ([[0, 1], [10, 11]], [[-1, 11], [0, 12]])  # for z  # for zT

    # standard  (lbi=True, ubi=False)
    assert (
        outRange(z, zT, rng=rng) == ~np.array([[True, False], [True, False]])
    ).all()

    # going through options
    assert (
        outRange(z, zT, rng=rng, lbi=False, ubi=False)
        == ~np.array([[False, False], [False, False]])
    ).all()
    assert (
        outRange(z, zT, rng=rng, lbi=False, ubi=True)
        == ~np.array([[False, True], [False, True]])
    ).all()
    assert (
        outRange(z, zT, rng=rng, lbi=True, ubi=False)
        == ~np.array([[True, False], [True, False]])
    ).all()
    assert (
        outRange(z, zT, rng=rng, lbi=True, ubi=True)
        == ~np.array([[True, True], [True, True]])
    ).all()

    return


# /def


def test_outRange_mixing_1D_and_ND_arguments():
    """Test outRange for a mix of 1D and N-D arguments.

    only mixed by breaking the ND arguments into a series of 1D arrays

    """
    # ---------------------------------
    # range agreement

    rng = (
        [0, 1],  # for x
        *[[0, 1], [10, 11]],  # for z
        [10, 11],  # for y
        *[[0, 10], [1, 11]],  # for zT
    )

    # standard  (lbi=True, ubi=False)
    assert (
        outRange(x, *z, y, *zT, rng=rng)
        == ~np.array([[True, False], [True, False]])
    ).all()

    # going through options
    assert (
        outRange(x, *z, y, *zT, rng=rng, lbi=False, ubi=False)
        == ~np.array([[False, False], [False, False]])
    ).all()
    assert (
        outRange(x, *z, y, *zT, rng=rng, lbi=False, ubi=True)
        == ~np.array([[False, True], [False, True]])
    ).all()
    assert (
        outRange(x, *z, y, *zT, rng=rng, lbi=True, ubi=False)
        == ~np.array([[True, False], [True, False]])
    ).all()
    assert (
        outRange(x, *z, y, *zT, rng=rng, lbi=True, ubi=True)
        == ~np.array([[True, True], [True, True]])
    ).all()

    return


# /def


##############################################################################
# ioRange


def test_ioRange_raises():
    """Test ioRange for expected failures."""
    with pytest.raises(ValueError):
        ioRange(incl=None, excl=None)


# /def


@pytest.mark.parametrize(
    "rng, expected",
    [
        (([0, 1], [11, 12]), np.array([True, False])),
        (([0, 1], [10, 11]), np.array([False, False])),
        (([1, 2], [11, 12]), np.array([False, False])),
    ],
    ids=["in & out", "both in", "both out"],
)
def test_ioRange_single_1D_argument(rng, expected):
    """Test ioRange for a single, 1D argument."""
    assert all(ioRange(incl=x, excl=y, rng=rng) == expected)


# /def


##############################################################################
# Ellipse


##############################################################################
# Circle


##############################################################################
# END



# -*- coding: utf-8 -*-

"""`lmfit <https://lmfit.github.io/lmfit-py/index.html>`_ Interface Utilities.

In particular, the `scipy_residual_to_lmfit` makes
`scipy <https://docs.scipy.org/doc/scipy/reference/optimize.html>`_ residuals
compatible with lmfit
`minimize <https://github.com/lmfit/lmfit-py/blob/master/lmfit/minimizer.py>`_.

"""


##############################################################################
# IMPORTS

# BUILT-IN
import typing as T

# THIRD PARTY
import numpy as np
from wrapt import ObjectProxy

try:
    import lmfit
except ImportError:
    HAS_LMFIT = False
else:
    HAS_LMFIT = True

##############################################################################
# PARAMETERS

if HAS_LMFIT:
    ParametersType = T.TypeVar("ParametersType", bound=lmfit.Parameters)
else:
    ParametersType = T.Any

##############################################################################
# CODE


class scipy_residual_to_lmfit(ObjectProxy):
    """Decorator to make scipy residual functions compatible with lmfit.

    (see https://lmfit.github.io/lmfit-py/fitting.html)

    Parameters
    ----------
    param_order : list of strs
        the variable order used by lmfit
        the strings are the names of the lmfit parameters
        must be in the same order as the scipy residual function

    Returns
    -------
    scipy_residual_to_lmfit : class
        internally constructed class

    Notes
    -----
    the function can be called as normal
    add a .lmfit function for use in lmfit minimizations
    see https://lmfit.github.io/lmfit-py/fitting.html

    >>> @scipy_residual_to_lmfit(param_order=['amp', 'phase', 'freq', 'decay'])
    ... def residual(variables, x, data, eps_data):
    ...     amp, phase, freq, decay = variables
    ...     # calculate residual here
    ...     return res

    .. todo::

        since using ObjectProxy, make it compatible with bound functions
        see https://wrapt.readthedocs.io/en/latest/wrappers.html

    """

    def __new__(
        cls,
        func: T.Callable = None,
        param_order: T.Optional[T.Sequence] = None,
    ):
        """Create Proxy."""
        if param_order is None:
            raise ValueError("param_order cannot be None")

        self = super().__new__(cls)  # inherit class information

        # assigning documentation as function documentation
        self.__doc__ = func.__doc__

        # allowing scipy_residual_to_lmfit to act as a decorator
        if func is None:
            return self.decorator(param_order)
        return self

    # /def

    @classmethod
    def decorator(cls, param_order: T.Sequence) -> T.Callable:
        """Decorator."""
        # @functools.wraps(cls)  # not needed when using ObjectProxy
        def wrapper(func: T.Callable) -> T.Callable:
            """scipy_residual_to_lmfit wrapper."""
            return cls(func, param_order=param_order)

        # /def
        return wrapper

    # /def

    def __init__(self, func: T.Callable, param_order: T.Sequence):
        """Initialize Proxy."""
        super().__init__(func)  # initializing function into ObjectProxy
        self.param_order = param_order
        return

    # /def

    def lmfit(
        self, params: ParametersType, *args: T.Any, **kwargs: T.Any
    ) -> T.Sequence:
        """`lmfit` version of function."""
        variables = [params[n].value for n in self.param_order]
        return self.__wrapped__(variables, *args, **kwargs)

    # /def

    def __call__(self, *args: T.Any, **kwargs: T.Any) -> T.Sequence:
        """Call scipy residual.

        Parameters
        ----------
        *args : Any
            Arguments into ``__wrapped__``, the residual function.
        **kwargs : Any
            Keyword arguments into ``__wrapped__``, the residual function.

        Returns
        -------
        return_ : Any
            Returns from called ``__wrapped__``.

        """
        return self.__wrapped__(*args, **kwargs)

    # /def


# /class


##############################################################################


def report_mcmc_fit(mcmc_res):
    """Report output of MCMC fit.

    Code from https://lmfit.github.io/lmfit-py/fitting.html

    """
    # -----------------------
    # Median of distribution

    print("median of posterior probability distribution")
    print("--------------------------------------------")
    lmfit.report_fit(mcmc_res.params)

    p = mcmc_res.params.copy()

    highest_prob = np.argmax(mcmc_res.lnprob)
    hp_loc = np.unravel_index(highest_prob, mcmc_res.lnprob.shape)
    mle_soln = mcmc_res.chain[hp_loc]
    for i, par in enumerate(p):
        p[par].value = mle_soln[i]

    # -----------------------
    # Max Likelihood

    print("\nMaximum Likelihood Estimation from emcee       ")
    print("-------------------------------------------------")
    print("Parameter  MLE Value   Median Value   Uncertainty")

    fmt = "  {:5s}  {:11.5f} {:11.5f}   {:11.5f}".format
    for name, param in p.items():
        print(
            fmt(
                name,
                param.value,
                mcmc_res.params[name].value,
                mcmc_res.params[name].stderr,
            )
        )

    # -----------------------
    # Error Estimate

    print("\nError Estimates from emcee    ")
    print("------------------------------------------------------")
    print("Parameter  -2sigma  -1sigma   median  +1sigma  +2sigma ")

    for name in p.keys():
        quantiles = np.percentile(
            mcmc_res.flatchain[name], [2.275, 15.865, 50, 84.135, 97.275]
        )
        median = quantiles[2]
        err_m2 = quantiles[0] - median
        err_m1 = quantiles[1] - median
        err_p1 = quantiles[3] - median
        err_p2 = quantiles[4] - median
        fmt = "  {:5s}   {:8.4f} {:8.4f} {:8.4f} {:8.4f} {:8.4f}".format
        print(fmt(name, err_m2, err_m1, median, err_p1, err_p2))

    # /for

    return


# /def


##############################################################################
# END



# -*- coding: utf-8 -*-

"""Test functions for astropy imports."""


__all__ = ["test_import_astropy"]


##############################################################################


def test_import_astropy():
    """Test _inRange."""
    from utilipy.imports import astropy_imports as imports

    for obj in (
        "astropy",
        "u",
        "coord",
        "SkyCoord",
        "Table",
        "QTable",
        "quantity_support",
        "astropy_mpl_style",
    ):
        assert hasattr(imports, obj)

    return


# /def


##############################################################################
# END



# -*- coding: utf-8 -*-

"""PDB and close-to-zero safe element-wise |Quantity| array comparisons.

The Astropy funciton :func:`~astropy.tests.helper.quantity_allclose` has
a few shortcomings:

    - It fails if the argument `a` and `b` are/close to zero
    - pdb diagnostics when the units don't match are really hard

"""

__all__ = [
    "eltwise_quantity_isclose",
    "eltwise_quantity_allclose",
    "eltwise_assert_quantity_isclose",
    "eltwise_assert_quantity_allclose",
]


##############################################################################
# IMPORTS

# BUILT-IN
import typing as T
from itertools import zip_longest

# THIRD PARTY
import astropy.units as u
import numpy as np
from astropy.units.quantity import (
    _unquantify_allclose_arguments as _unquantify,
)

##############################################################################
# PARAMETERS

QuantityType = T.TypeVar("QuantityType", bound=u.Quantity)

##############################################################################
# CODE
##############################################################################


# TODO support argument kwargs
def eltwise_quantity_isclose(
    a,
    b,
    rtol=1e-15,
    atol=None,
    equal_nan=False,
    wrap: T.Union[None, T.Tuple[int, QuantityType]] = None,
):
    """Returns True if two arrays are element-wise equal within a tolerance.

    This is a |Quantity|-aware version of :func:`~numpy.allclose`,
    modified from :mod:`~astropy` to be easier for PDB debugging.

    .. warning::

        This function should only be used when setting up testing.
        Use :func:`~astropy.units.allclose` or
        :func:`~astropy.tests.helper.quantity_allclose`.

    Parameters
    ----------
    wrap : T.Tuple[int, QuantityType], optional

    """
    # Splitting the comparison into a for-loop allows for element-wise
    # comparisons and prevents dropping into sub-functions so we stay in this
    # namespace when in PDB.
    try:
        alen = len(a)
    except TypeError:  # scalar
        a = [a]
        alen = 1

    try:
        blen = len(b)
    except TypeError:  # scalar
        b = [b]
        blen = 1

    assert alen == blen

    try:
        len(rtol)
    except TypeError:  # scalar
        rtol = np.broadcast_to(rtol, blen, subok=True)

    try:
        len(atol)
    except TypeError:  # scalar
        atol = np.broadcast_to(atol, blen, subok=True)

    wrap = wrap if wrap is not None else [wrap]

    close = []
    for x, y, rt, at, wrp in zip_longest(a, b, rtol, atol, wrap):
        if wrp is not None:  # adjust to phase-wrap
            x = np.divmod(x, wrp)[1]  # the "remainder"
            y = np.divmod(y, wrp)[1]

        try:
            x, y, _rt, _at = _unquantify(x, y, rt, at)
        except u.UnitsError as e:
            raise u.UnitsError(e)

        compare = u.isclose(x, y, rtol=_rt, atol=_at, equal_nan=equal_nan)

        close.append(compare)

    return np.array(close)


# /def


# -------------------------------------------------------------------


# TODO support argument kwargs
def eltwise_quantity_allclose(
    a,
    b,
    rtol=1e-15,
    atol=None,
    equal_nan=False,
    wrap: T.Union[None, T.Tuple[int, QuantityType]] = None,
):
    """Returns True if two arrays are element-wise equal within a tolerance.

    This is a |Quantity|-aware version of :func:`~numpy.allclose`,
    modified from :mod:`~astropy` to be easier for PDB debugging.

    .. warning::

        This function should only be used when setting up testing.
        Use :func:`~astropy.units.allclose` or
        :func:`~astropy.tests.helper.quantity_allclose`.

    """
    return np.all(
        eltwise_quantity_isclose(
            a, b, rtol=rtol, atol=atol, equal_nan=equal_nan, wrap=wrap
        )
    )


# /def


# -------------------------------------------------------------------


# TODO support argument kwargs
def eltwise_assert_quantity_isclose(
    a,
    b,
    rtol=1e-15,
    atol=None,
    equal_nan=False,
    wrap: T.Union[None, T.Tuple[int, QuantityType]] = None,
):
    """Returns True if two arrays are element-wise equal within a tolerance.

    This is a |Quantity|-aware version of :func:`~numpy.allclose`,
    modified from :mod:`~astropy` to be easier for PDB debugging.

    .. warning::

        This function should only be used when setting up testing.
        Use :func:`~astropy.units.allclose` or
        :func:`~astropy.tests.helper.quantity_allclose`.

    """
    # Splitting the comparison into a for-loop allows for element-wise
    # comparisons and prevents dropping into sub-functions so we stay in this
    # namespace when in PDB.
    try:
        alen = len(a)
    except TypeError:  # scalar
        a = [a]
        alen = 1

    try:
        blen = len(b)
    except TypeError:
        b = [b]
        blen = 1

    assert alen == blen

    try:
        len(rtol)
    except TypeError:  # scalar
        rtol = np.broadcast_to(rtol, blen, subok=True)

    try:
        len(atol)
    except TypeError:  # scalar
        atol = np.broadcast_to(atol, blen, subok=True)

    wrap = wrap if wrap is not None else [wrap]

    for x, y, rt, at, wrp in zip_longest(a, b, rtol, atol, wrap):
        if wrp is not None:  # adjust to phase-wrap
            x = np.divmod(x, wrp)[1]  # the "remainder"
            y = np.divmod(y, wrp)[1]

        try:
            x, y, _rt, _at = _unquantify(x, y, rt, at)
        except u.UnitsError as e:
            raise u.UnitsError(e)

        assert u.isclose(
            x, y, rtol=_rt, atol=_at, equal_nan=equal_nan
        ), f"{x}, {y} | {_rt}, {_at}"


# /def


# -------------------------------------------------------------------


# TODO support argument kwargs
def eltwise_assert_quantity_allclose(
    a,
    b,
    rtol=1e-15,
    atol=None,
    equal_nan=False,
    wrap: T.Union[None, T.Tuple[int, QuantityType]] = None,
):
    """Raise an assertion if two objects are not equal up to desired tolerance.

    This is a :class:`~astropy.units.Quantity`-aware version of
    :func:`numpy.testing.assert_allclose`,
    modified from :mod:`~astropy` to be easier for PDB debugging.

    .. warning::

        This function should only be used when setting up testing.
        Use :func:`~astropy.tests.helper.assert_quantity_allclose`.

    """
    # Splitting the comparison into a for-loop allows for element-wise
    # comparisons and prevents dropping into sub-functions so we stay in this
    # namespace when in PDB.
    return eltwise_assert_quantity_isclose(
        a, b, rtol=rtol, atol=atol, equal_nan=equal_nan, wrap=wrap
    )


# /def


##############################################################################
# END



# -*- coding: utf-8 -*-

# ----------------------------------------------------------------------------
#
# TITLE   : matplotlib_imports
# PROJECT : utilipy
#
# ----------------------------------------------------------------------------

"""Base set of imports for :mod:`~matplotlib`.

Routine Listings
----------------
Matplotlib: imports

    - pyplot->plt
    - matplotlib->mpl, .cm, .colors
    - mpl_toolkits.mplot3d.Axes3D

utilipy: imports

    - ipython.plot.configure_matplotlib

References
----------
Matplotlib reference [#]_.

.. [#] John D. Hunter. Matplotlib: A 2D Graphics Environment, Computing in
    Science and Engineering, 9, 90-95 (2007), DOI:10.1109/MCSE.2007.55
    http://scitation.aip.org/content/aip/journal/cise/9/3/10.1109/MCSE.2007.55

"""

__author__ = "Nathaniel Starkman"


__all__ = [
    "mpl",
    "plt",
    "cm",
    "colors",
    # "Axes3D",
    "matplotlib_imports_help",
]


##############################################################################
# HELPER FUNCTIONS

# THIRD PARTY
import matplotlib as mpl
import matplotlib.pyplot as plt
from matplotlib import cm, colors

# PROJECT-SPECIFIC
from utilipy.imports import conf
from utilipy.utils import make_help_function

# from mpl_toolkits.mplot3d import Axes3D


# +--------------------------------------------------------------------------+
# IPython Magic

try:

    _HAS_IPYTHON = False

    get_ipython()

    if get_ipython() is None:  # double checking
        raise NameError

except NameError:

    pass

else:

    _HAS_IPYTHON = True

    from utilipy.ipython.plot import configure_matplotlib

    __all__ += [
        "configure_matplotlib",
    ]  # TODO add when no IPython


##############################################################################
# Running Imported Functions

if _HAS_IPYTHON:

    configure_matplotlib(backend="inline", figure_format="retina")


##############################################################################
# Printing Information

matplotlib_imports_help = make_help_function(
    "matplotlib", __doc__, look_for="Routine Listings"
)


if conf.verbose_imports:
    matplotlib_imports_help()


##############################################################################
# END



# -*- coding: utf-8 -*-

# ----------------------------------------------------------------------------
#
# TITLE   : Napoleon Parse tools
# PROJECT : utilipy
#
# ----------------------------------------------------------------------------

# Docstring
"""Napoleon parse tools.

routines to parse docstrings according to the Napoleon format specifications.
code sourced from [1]_, but modified to merge, not replace, docstring sections.


Routine Listings
----------------
merge_google_napoleon_docs
    Merge two google-style docstrings into a single docstring,
    according to napoleon docstring sections.

merge_numpy_napoleon_docs
    Merge two numpy-style docstrings into a single docstring,
    according to napoleon docstring sections.

References
----------
The code source [1]_.

.. [1] https://github.com/rsokl/custom_inherit

"""

__author__ = "Ryan Soklaski"
__license__ = "MIT"
__version__ = "2.2.2"
__email__ = "https://github.com/rsokl/custom_inherit"

__all__ = ["merge_google_napoleon_docs", "merge_numpy_napoleon_docs"]


###############################################################################
# IMPORTS

# BUILT-IN
import typing as T
from collections import OrderedDict
from inspect import cleandoc

# THIRD PARTY
from typing_extensions import Literal

###############################################################################
# CODE
###############################################################################


def parse_napoleon_doc(doc: T.Union[str, None], style: str):
    """Extract the text from the sections of a numpy-formatted docstring.

    Parameters
    ----------
    doc: str or None
        The docstring to parse.

    style: str
        'google' or 'numpy'

    Returns
    -------
    OrderedDict[str, Union[None,str]]
        The extracted numpy-styled docstring sections.

    """
    napoleon_sections = [
        "Short Summary",
        "Attributes",
        "Methods",
        "Warning",
        "Note",
        "Parameters",
        "Other Parameters",
        "Keyword Arguments",
        "Returns",
        "Yields",
        "Raises",
        "Warns",
        "See Also",
        "References",
        "Todo",
        "Example",
        "Examples",
    ]

    aliases = {
        "Args": "Parameters",
        "Arguments": "Parameters",
        "Keyword Args": "Keyword Arguments",
        "Return": "Returns",
        "Warnings": "Warning",
        "Yield": "Yields",
    }

    doc_sections = OrderedDict([(key, None) for key in napoleon_sections])

    if not doc:
        return doc_sections

    assert style in ("google", "numpy")

    doc = cleandoc(doc)
    lines = iter(doc.splitlines())

    key = "Short Summary"
    body = []
    while True:
        try:
            line = next(lines).rstrip()
            header = (
                line
                if style == "numpy"
                else (line[:-1] if line.endswith(":") else line)
            )
            if header and (header in doc_sections or header in aliases):
                doc_sections[aliases.get(key, key)] = (
                    "\n".join(body).rstrip() if body else None
                )
                body = []
                key = header
                if style == "numpy":
                    next(lines)  # skip section delimiter
            else:
                body.append(line)
        except StopIteration:
            doc_sections[aliases.get(key, key)] = "\n".join(body)
            break
    return doc_sections


# /def


def merge_section(
    key: str,
    prnt_sec: T.Union[str, None],
    child_sec: T.Union[str, None],
    style: str,
    method: T.Union[Literal["merge"], Literal["replace"]] = "replace",
):
    """Synthesize a output napoleon docstring section.

    Parameters
    ----------
    key: str
        The napoleon-section being merged.
    prnt_sec: Union[str, None]
        The docstring section from the parent's attribute.
    child_sec: Union[str, None]
        The docstring section from the child's attribute.
    style: str
        docstring style format
    method: {merge, replace}, optional
        'merge' merges the sections.
        'replace' (default) parent with child sections, if exists.

    Returns
    -------
    Optional[str]
        The output docstring section.

    """
    if prnt_sec is None and child_sec is None:
        return None

    assert style in ("google", "numpy")

    if key == "Short Summary":
        header = ""
    else:
        if style == "numpy":
            header = "\n".join(
                (key, "".join("-" for i in range(len(key))), "")
            )
        else:
            header = "\n".join((key + ":", ""))

    if method == "replace":
        body = prnt_sec if child_sec is None else child_sec
    elif method == "merge":
        # TODO more intelligent merge strategy,
        # respecting arg vs kwarg order, overriding repeated args, etc.
        body = prnt_sec if child_sec is None else prnt_sec + "\n" + child_sec

    return header + body


# /def


def merge_all_sections(
    prnt_sctns: T.MutableMapping[str, T.Union[None, str]],
    child_sctns: T.MutableMapping[str, T.Union[None, str]],
    style: str,
    method: T.Union[Literal["merge"], Literal["replace"]] = "replace",
):
    """Merge the doc-sections of the parent's and child's attribute.

    Parameters
    ----------
    prnt_sctns: OrderedDict[str, Union[None,str]]
    child_sctns: OrderedDict[str, Union[None,str]]
    method: {merge, replace}, optional
        'merge' merges the sections.
        'replace' (default) parent with child sections, if exists.

    Returns
    -------
    str
        Output docstring of the merged docstrings.

    """
    doc = []

    prnt_only_raises = prnt_sctns["Raises"] and not (
        prnt_sctns["Returns"] or prnt_sctns["Yields"]
    )
    if prnt_only_raises and (child_sctns["Returns"] or child_sctns["Yields"]):
        prnt_sctns["Raises"] = None

    for key in prnt_sctns:
        sect = merge_section(
            key, prnt_sctns[key], child_sctns[key], style, method=method
        )
        if sect is not None:
            doc.append(sect)
    return "\n\n".join(doc) if doc else None


# /def


def merge_numpy_napoleon_docs(
    prnt_doc: T.Optional[str] = None,
    child_doc: T.Optional[str] = None,
    method: T.Union[Literal["merge"], Literal["replace"]] = "replace",
):
    """Merge two numpy-style docstrings into one numpy-napolean docstring.

    Given the numpy-style docstrings from a parent and child's attributes,
    merge the docstring sections such that the child's section is used,
    wherever present, otherwise the parent's section is used.

    Any whitespace that can be uniformly removed from a docstring's second line
    and onwards is removed. Sections will be separated by a single blank line.

    Aliased docstring sections are normalized.
    E.g Args, Arguments -> Parameters

    Parameters
    ----------
    prnt_doc: Optional[str]
        The docstring from the parent.
    child_doc: Optional[str]
        The docstring from the child.
    method: {merge, replace}, optional
        'merge' merges the sections.
        'replace' (default) parent with child sections, if exists.

    Returns
    -------
    Union[str, None]
        The merged docstring.

    """
    style = "numpy"
    return merge_all_sections(
        parse_napoleon_doc(prnt_doc, style),
        parse_napoleon_doc(child_doc, style),
        style,
        method=method,
    )


# /def


def merge_google_napoleon_docs(
    prnt_doc: T.Optional[str] = None,
    child_doc: T.Optional[str] = None,
    method: T.Union[Literal["merge"], Literal["replace"]] = "replace",
):
    """Merge two google-style docstrings into one google-napoleon docstring.

    Given the google-style docstrings from a parent and child's attributes,
    merge the docstring sections such that the child's section is used,
    wherever present, otherwise the parent's section is used.

    Any whitespace that can be uniformly removed from a docstring's second line
    and onwards is removed. Sections will be separated by a single blank line.

    Aliased docstring sections are normalized.
    E.g Args, Arguments -> Parameters

    Parameters
    ----------
    prnt_doc: Optional[str]
        The docstring from the parent.
    child_doc: Optional[str]
        The docstring from the child.
    method: {merge, replace}, optional
        'merge' merges the sections.
        'replace' (default) parent with child sections, if exists.

    Returns
    -------
    Union[str, None]
        The merged docstring.

    """
    style = "google"
    return merge_all_sections(
        parse_napoleon_doc(prnt_doc, style),
        parse_napoleon_doc(child_doc, style),
        style,
        method=method,
    )


# /def

###############################################################################
# END



# -*- coding: utf-8 -*-

"""Base class for decorators.

Most decorators are function-based. A function-based decorator is reasonably
flexible, even permitting default values to be set dynamically. By using
`~utilipy.utils.functools.wraps` in :mod:`~utility.utils.functools` these
function-based decorators may can also merge the decorator and function
signatures and docstrings.

.. code-block:: python

    def template_decorator(function=None, *, kw1=None, kw2=None):
        '''Decorator Docstring.

        Parameters
        ----------
        function : Callable or None, optional
            the function to be decoratored
            if None, then returns decorator to apply.
        kw1, kw2 : Any, optional
            keyword only arguments
            sets the wrapper's default values.

        Returns
        -------
        wrapper : Callable
            wrapper for `function`
            includes the original function in a method ``.__wrapped__``

        '''
        if function is None: # allowing for optional arguments
            return functools.partial(template_decorator, kw1=k1, kw2=kw2)

        @utilipy.wraps(function, _doc_style='numpy')
        def wrapper(*args, kw1=kw1, kw2=kw2, kw3="kw3", **kw):
            '''wrapper docstring.

            Parameters
            ----------
            kw1, kw2 : Any, optional
                default {kw1}, {kw2}
            kw3 : str, optional
                default {kw3}

            '''
            # do stuff here
            return_ = function(*args, **kw)
            # and here
            return return_

        return wrapper

For most purposes decorators of these sorts are sufficient. However, there are
a number of limitations inherent to a function-based approach. A more general
architecture is possible using classes. As example, ``TemplateDecorator``
is nearly equivalent to the previously defined ``template_decorator``.

.. code-block:: python

    class TemplateDecorator():

        def __new__(cls, function=None, *, kw1=None, kw2=None):
            self = super().__new__(cls)  # make instance
            if function is not None:  # this will return a wrapped function
                return self(function, kw1=kw1, kw2=kw2)
            else:  # this will return a function wrapper
                return self

        def __init__(self, kw1=None, kw2=None):
            self.kw1 = kw1
            self.kw2 = kw2

        def __call__(self, function, kw1=None, kw2=None):
            '''Construct a function wrapper.'''
            kw1 = self.kw1 if kw1 is None else kw1
            kw2 = self.kw2 if kw2 is None else kw2

            @utilipy.wraps(function)
            def wrapper(*args, kw1=kw1, kw2=kw2, kw3="kw3", **kw):
                '''wrapper docstring.

                Parameters
                ----------
                kw1, kw2 : Any, optional
                    default {kw1}, {kw2}
                kw3 : str, optional
                    default {kw3}

                '''
                # do stuff here
                return_ = function(*args, **kw)
                # and here
                return return_

            return wrapper

An important difference between ``template_decorator`` and
``TemplateDecorator`` happens at creation. If template decorator is
created with


"""

__author__ = "Nathaniel Starkman"


__all__ = [
    # decorators
    "DecoratorBaseMeta",
    "DecoratorBaseClass",
    "classy_decorator",
    # utility
    "_placeholders",
]


##############################################################################
# IMPORTS

# BUILT-IN
import copy
import typing as T
from abc import abstractmethod
from collections import namedtuple

# PROJECT-SPECIFIC
from utilipy.utils import functools, inspect
from utilipy.utils.string import FormatTemplate

##############################################################################
# Parameters

_placeholders: tuple = (inspect._empty, inspect._void, inspect._placehold)


class kwargs:  # FIXME: temporary fix for Sphinx
    pass


##############################################################################
# CODE
##############################################################################

# ----------------------------------------------------------------------------
# MetaClass


class DecoratorBaseMeta(type):
    """Meta-class for decorators.

    This metaclass is pretty specific for DecoratorBaseClass and should
    probably not be used in other contexts.

    This metaclass:

    - creates `__kwdefaults__`, copying from `__base_kwdefaults__`.
    - makes a named-tuple class for read-only versions of `__kwdefaults__`
    - stores the original class and call docs
    - formats the class docstrings with the `__kwdefaults__`

    """

    def __new__(cls: type, name: str, bases: tuple, dct: dict):
        """Set properties for new decorator class.

        define an `__init__` method and store original docs

        Parameters
        ----------
        name : str
        bases : tuple
        dct : dictionary

        """
        # --------------------------------------------------
        # __base_kwdefaults__ -> __kwdefaults__

        dct["__kwdefaults__"] = dct["__base_kwdefaults__"].copy()

        dct["_kwd_namedtuple"] = namedtuple(
            "kwdefaults", dct["__kwdefaults__"].keys()
        )

        # --------------------------------------------------
        # store original docs
        # with None -> ''
        dct["_orig_classdoc_"] = FormatTemplate(
            copy.copy(dct.get("__doc__", None)) or ""
        )
        dct["_orig_calldoc_"] = FormatTemplate(
            copy.copy(dct["__call__"].__doc__) or ""
        )

        # formatting class docstring
        # needs to be done here for the first class declaration
        # later edits are handled in DecoratorBaseClass by `__new__`
        dct["__doc__"]: str = dct["_orig_classdoc_"].safe_substitute(
            **dct["__kwdefaults__"]
        )

        # ------------------------
        # 1) change call's signature
        #   - add a variable positional argument (\*args)
        #   - promote any positional arguments with defaults
        #     (except *function*) to keyword only arguments
        #   - substitue the values of `__kwdefaults__` as the default values
        #     for key-word arguments with a placehold value
        #     (_empty, _void, _placehold)

        callsig: inspect.FullerSignature
        callsig = inspect.fuller_signature(dct["__call__"])

        # store whether has kwargs, for use in ``callwrapper``
        has_kwargs = callsig.index_var_keyword

        # next check if have a var_positional argument
        # promoting default-valued positional arguments to kwargs
        callsig = callsig.add_var_positional_parameter(index=2)
        callsig = callsig.add_var_keyword_parameter(name="kwargs")

        # replacing defaults if in __base_kwdefaults__
        i: int
        name: str
        param: inspect.Parameter
        for i, (name, param) in enumerate(callsig.parameters.items()):
            # only keyword-only args (most since added var_positional)
            if (name in dct["__kwdefaults__"]) and (  # yup
                param.kind == inspect.KEYWORD_ONLY
            ):  # key-word
                callsig = callsig.modify_parameter(
                    i, default=dct["__kwdefaults__"][name]
                )

        # updating __call__ from callsig
        attr: str
        for attr in functools.SIGNATURE_ASSIGNMENTS:
            try:
                value = getattr(callsig, attr)
            except AttributeError:
                pass
            else:
                setattr(dct["__call__"], attr, value)

        # ------------------------
        # 2) Wrap __call__ so can use without parenthesis
        #   - store __call__
        #   - change function default to None

        # store original call function   # TODO fix copy_function
        dct["__orig_call__"]: T.Callable = functools.copy_function(
            dct["__call__"]
        )
        dct["__orig_call__"].__kwdefaults__: dict = dct[
            "__call__"
        ].__kwdefaults__
        # dct["__orig_call__"].__signature__ = callsig.signature

        # change function default to None
        # not doing it before to preserve __orig_call__
        try:
            callsig = callsig.modify_parameter(
                1, default=None, kind=inspect.POSITIONAL_OR_KEYWORD
            )
        except ValueError:  # already has a default
            pass

        dct["_callsig"]: inspect.Signature = callsig

        # make wrapper
        def callwrapper(
            self, function: T.Callable = None, *args: T.Any, **kwargs: T.Any
        ):
            self = self.new(
                **kwargs
            )  # TODO extract __kwdefaults__ dict from this

            # update kwargs to carry over to __orig_call__
            if has_kwargs is False:  # no kwargs, can only use existing
                nkw: dict = (self._callsig.__kwdefaults__ or {}).copy()
            else:  # can pass everything along
                nkw: dict = (self.__kwdefaults__ or {}).copy()
            nkw.update(kwargs)

            # allow calling without parenthesis
            if function is None:
                return functools.partial(self.__orig_call__, **nkw)
            else:
                return self.__orig_call__(function, *args, **nkw)

        # /def

        callwrapper.__kwdefaults__: dict = dct["__kwdefaults__"].copy()
        # /def

        # overwrite __call__ with wrapped function
        # TODO fix update_wrapper
        dct["__call__"]: T.Callable = functools.update_wrapper(
            callwrapper,
            dct["__orig_call__"],
            signature=callsig,
            docstring=dct["_orig_calldoc_"].safe_substitute(
                **dct["__kwdefaults__"]
            ),
        )
        dct["__call__"].__kwdefaults__: dict = dct["__kwdefaults__"].copy()

        # ------------------------

        # create and return object
        return type.__new__(cls, name, bases, dct)

    # /def

    def __init__(cls: type, name: str, bases: tuple, dct: dict):
        """__init__ method for MetaClass.

        Sets up docstring inheritance.

        References
        ----------
        The docstring inheritance method is modified from
        astropy's InheritDocstrings metaclass

        .. todo::

            replace with better doc_inheritance from custom_inherit

        """
        # ---------------------
        # docstring inheritance

        # find public methods
        def is_public_member(key: str):
            return (
                key.startswith("__") and key.endswith("__") and len(key) > 4
            ) or not key.startswith("_")

        # /def

        # add docstring to methods
        key: str
        val: T.Any
        for key, val in dct.items():
            # check is public function without a docstring
            if (
                (inspect.isfunction(val) or inspect.isdatadescriptor(val))
                and is_public_member(key)
                and val.__doc__ is None
            ):

                # traverse MRO
                for base in cls.__mro__[1:]:
                    super_method = getattr(base, key, None)
                    if super_method is not None:
                        val.__doc__ = super_method.__doc__
                        break
                # /for
        # /for

        # ---------------------

        super().__init__(name, bases, dct)

        return

    # /def


# /class


# ----------------------------------------------------------------------------
# Base Class


class DecoratorBaseClass(metaclass=DecoratorBaseMeta):
    """Base Class for Decorators.

    Parameters
    ----------
    function : Callable, optional
    kwargs : dict, optional
        attributes

    Returns
    -------
    type
        function if called with the function (Decorator(function))
        type if no function provided, i.e. pie syntax

    Raises
    ------
    ValueError
        if a kwarg not supported, i.e. not in __kwdefaults__

    """

    __base_kwdefaults__: dict = {}

    @property
    def kwdefaults(self: type):
        """__kwdefaults__ named tuple."""
        return self._kwd_namedtuple(**self.__kwdefaults__)

    # /def

    def __new__(
        cls: type, function: T.Optional[T.Callable] = None, **kwargs: T.Any
    ):
        """Make new DecoratorBaseClass.

        Parameters
        ----------
        function : Collable, optional
        kwargs : dict
            attributes

        Returns
        -------
        type
            function if called with the function (Decorator(function))
            type if no function provided, i.e. pie syntax

        Raises
        ------
        ValueError
            if a kwarg not supported, i.e. not in __kwdefaults__

        """
        self = super().__new__(cls)  # make instance

        # --------------------
        # assign to class

        attr_keys = self.__kwdefaults__.keys()  # get defaults

        # check keys are valid
        key: str
        for key in kwargs.keys():
            if key not in attr_keys:
                raise ValueError("")

        # update __kwdefaults__ from kwargs
        self.__kwdefaults__ = self.__base_kwdefaults__.copy()
        self.__kwdefaults__.update(kwargs)

        # --------------------
        # format docstrings
        # done here for all the non-first class creation (see metaclass)

        self.__doc__: str = self._orig_classdoc_.safe_substitute(
            **self.__kwdefaults__
        )

        self.__call__.__func__.__doc__: str
        self.__call__.__func__.__doc__ = self._orig_calldoc_.safe_substitute(
            **self.__kwdefaults__
        )

        # --------------------
        # format __call__ signature

        # replacing defaults if in __kwdefaults__
        for i, (name, param) in enumerate(self._callsig.parameters.items()):
            # only keyword-only args (most since added var_positional)
            if (
                (name in self.__kwdefaults__)  # yup
                and (param.kind == inspect.KEYWORD_ONLY)  # key-word
                # and (param.default in _placeholders)  # no default!
            ):
                self._callsig = self._callsig.modify_parameter(
                    i, default=self.__kwdefaults__[name]
                )
        # /for

        # set signature and defaults
        self.__call__.__func__.__signature__ = self._callsig.signature
        self.__call__.__func__.__kwdefaults__ = self.__kwdefaults__

        # print(self._callsig.parameters.values())
        # print(self.__call__.__func__.__signature__)
        # print(self.__call__.__func__.__kwdefaults__)

        # --------------------

        if function is not None:  # this will return a wrapped function
            return self(function)
        else:  # this will return a function wrapper
            return self

    # /def

    def __getitem__(self: type, name: str):
        """Get item from kwdefaults."""
        return self.__kwdefaults__[name]

    # /def

    def __setitem__(self: type, name: str, value: T.Any):
        """Set kwdefaults item."""
        self.__kwdefaults__[name] = value

    # /def

    def as_decorator(
        self: type,
        wrapped_function: T.Optional[T.Callable] = None,
        **kwdefaults: T.Any
    ):
        """Make decorator.

        Parameters
        ----------
        wrapped_function : Callable
        kwdefaults
            update the decorator default values

        Returns
        -------
        Callable or type
            function if called with the function (Decorator(function))
            type if no function provided, i.e. pie syntax

        """
        decorator: type = self.new(**kwdefaults)

        if wrapped_function is not None:  # return a wrapped function
            return decorator(wrapped_function)
        # else return a function wrapper
        return decorator

    # /def

    def new(self: type, **kw):
        """Make a new decorator from current decorator.

        Inherits properties from current decorator.
        Can be overwritten.

        Parameters
        ----------
        kw: dict
            key, value pairs for the kwdefaults

        Returns
        -------
        type
            new decorator with updated properties

        TODO
        ----
        figure out how help(self) can correctly display __kwdefaults__

        """
        # get and update defaults
        kwdefaults: dict = self.__kwdefaults__.copy()
        kwdefaults.update(**kw)
        # make new class, updating defaults
        # TODO use new so override class docstring and recall metaclass
        return self.__class__(function=None, **kwdefaults)

    # /def

    @abstractmethod
    def __call__(self: type, function: T.Callable):
        @functools.wraps(function)
        def wrapper(*func_args: T.Any, **func_kwargs: T.Any):
            return function(*func_args, **func_kwargs)

        # /def

        return wrapper

    # /def


# /class


# ----------------------------------------------------------------------------
# Turn a Function into a Decorator


def classy_decorator(decorator_function: T.Callable = None):
    """Convert decorated functions to classes.

    Parameters
    ----------
    decorator_function: Callable
        The decorator function

    Returns
    -------
    Decorator: DecoratorBaseClass
        class version of `decorator_function`, using the `DecoratorBaseClass`

    """
    if decorator_function is None:  # allowing for optional arguments
        return functools.partial(classy_decorator)

    signature: inspect.FullerSignature
    signature = inspect.fuller_signature(decorator_function)
    signature = signature.prepend_parameter(  # add in self
        inspect.Parameter("self", inspect.POSITIONAL_ONLY)
    )
    # key-word defaults from function
    kwd: dict = inspect.getfullerargspec(decorator_function).kwonlydefaults

    def mycall(
        self: type, wrapped_function: T.Callable, *args: T.Any, **kwargs: T.Any
    ):
        return decorator_function(wrapped_function, *args, **kwargs)

    # /def

    class Decorator(DecoratorBaseClass):

        __base_kwdefaults__: dict = kwd

        # /def
        __call__: T.Callable = functools.makeFunction(
            mycall.__code__,
            mycall.__globals__,
            name="__call__",
            signature=signature,
            docstring=decorator_function.__doc__,
            closure=mycall.__closure__,
        )

    # /class

    # set docstring
    Decorator.__doc__: str = FormatTemplate(
        (decorator_function.__doc__ or "")
    ).safe_substitute(**kwd)
    Decorator.__name__ = decorator_function.__name__

    return Decorator()


# /def


##############################################################################
# END
